/* mirror.c
 * Copyright (C) 2006 Tillmann Werner <tillmann.werner@gmx.de>
 *
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without
 * modifications, as long as this notice is preserved.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <errno.h>

#include "mirror.h"
#include "logging.h"
#include "tcpserver.h"

int mirror_connect(struct in_addr ipaddr, u_int16_t port) {
	int mirror_sock_fd, local_addr_len;
	struct sockaddr_in mirror_socket, local_socket;
	
	/* prevent loops - disallow mirrors to localhost */
	if (ntohl(ipaddr.s_addr) == 0x7F000001) {
		logmsg(LOG_WARN, 1,
			"Connection Mirror warning - Mirroring to %s:%u suppressed for loop prevention.\n",
			inet_ntoa(ipaddr.s_addr), ntohs(port));
		return(-1);
	}

	/* prepare client socket */

	logmsg(LOG_DEBUG, 1, "Connection Mirror - Creating client socket.\n");
	if (!(mirror_sock_fd = socket(AF_INET, SOCK_STREAM, 0))) { 
		logmsg(LOG_ERR, 1, "Connection Mirror error - Unable to create client socket.\n");
		return(-1);
	} else {
		logmsg(LOG_DEBUG, 1, "Connection Mirror - Client socket created.\n");

		/* establish mirror connection */
		logmsg(LOG_DEBUG, 1, "Connection Mirror - Connecting to %s:%u.\n",
			inet_ntoa(ipaddr.s_addr), ntohs(port));

		bzero(&mirror_socket, sizeof(mirror_socket));
		mirror_socket.sin_family	= AF_INET;
		mirror_socket.sin_addr.s_addr	= ipaddr.s_addr;
		mirror_socket.sin_port		= port;
		
		if (connect(mirror_sock_fd, (struct sockaddr *) &mirror_socket, sizeof(mirror_socket)) != 0) {
			close(mirror_sock_fd);
			logmsg(LOG_DEBUG, 1, "Connection Mirror error - Unable to connect to %s:%d.\n",
				inet_ntoa(ipaddr.s_addr), ntohs(port));
			return(-1);
		}

		local_addr_len = 0;
		if (getsockname(mirror_sock_fd, (struct sockaddr *) &local_socket, &local_addr_len) != 0) {
			logmsg(LOG_ERR, 1, "Error - Unable to get local address from mirror socket: %s\n",
				strerror(errno));
			return(-1);
		}
		attack.m_conn.l_addr	= local_socket.sin_addr; 
		attack.m_conn.r_addr	= mirror_socket.sin_addr;
		attack.m_conn.l_port	= local_socket.sin_port;
		attack.m_conn.r_port	= mirror_socket.sin_port;
	}
	return(mirror_sock_fd);
}

int mirror_data(int mirror_sock_fd, u_char* buffer, size_t bytes_to_write) {
	ssize_t bytes_written;

	bytes_written = 0;

	logmsg(LOG_NOISY, 1, "Connection Mirror - Mirroring %u bytes.\n", bytes_to_write);

	if ((bytes_written = write(mirror_sock_fd, buffer, bytes_to_write)) == -1) { 
		logmsg(LOG_ERR, 1, "Connection Mirror error - Sending data failed, closing connection.\n");
		mirror_connected = 0;
		close(mirror_sock_fd);
		return(-1);
	}
	if (bytes_written != bytes_to_write)
		logmsg(LOG_WARN, 1, "Connection Mirror warning - %u of %u bytes sent to remote host.\n",
			bytes_written, bytes_to_write);

	return(bytes_written);
}
