/* readconf.c
 * Copyright (C) 2006 Tillmann Werner <tillmann.werner@gmx.de>
 *
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without
 * modifications, as long as this notice is preserved.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <errno.h>
#include <stdio.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <unistd.h>
#include <netdb.h>

#include "readconf.h"
#include "honeytrap.h"
#include "logging.h"
#include "response.h"
#include "interface.h"

#define OPT_IS(A)	(strcmp(low_opt, (A)) == 0)
#define OPTSTRING	"t:vh?pDmL:P:C:i:a:l:r:u:g:"


char *get_next_line(FILE * file) {
	/* return next line from file */

	char *recvline;
	char buf[BUFSIZ];
	char *index;

	bzero((char *)buf, BUFSIZ);

	while(fgets(buf, BUFSIZ, file)) {
		index = buf;
		/* advance through whitespaces at the beginning of the line */
		while (isspace((int) *index)) ++index;

		return((char *) strdup(index));
	}
	return(NULL);
}


void *get_value(char *buf, const char delim) {
	/* find delimiter in string and overwrite it with \0 to terminate keywort,
	 * skip following whitespaces and return pointer to the found value
	 * return NULL if delimiter is not found */

	char *index, *retval = NULL;

	/* search for delimiter */
	if ((retval = strchr(buf, delim)) == NULL) return(NULL);

	/* overwrite delimiter to terminate keyword */
	retval[0] = '\0';
	retval++;
	index = buf;
	while (!isspace((int) *index)) index++;
	index[0] = '\0';

	/* advance through whitespaces */
	while (isspace((int) *retval)) retval++;

	return(retval);
}


int configure(int my_argc, char *my_argv[]) {
	char *bpf_cmd_ext, option;
	struct hostent *ip_cmd_opt;
	struct if_data *interface;
	struct sockaddr *sa;
	struct sockaddr_in *sa_in;
	struct passwd *pwd_entry;
	struct group *grp_entry;

	/* initialilzation of variables */
	default_response = NULL;	/* initialize default response list */
	dev = NULL;			/* device pointer */
	ip_cmd_opt = NULL;		/* source ip address for bpf string */
	bpf_cmd_ext = NULL;		/* bpf string extension ('expression' from command line)
	pidfile_fd = 0;


fprintf(stderr, "---> Processing command line options.\n");
	/* scan command line options to determine logging level or print version number or usage */
	while((option = getopt(my_argc, my_argv, OPTSTRING)) > 0) {
		switch(option) {
			case 't':
				if ((atoi(optarg) < 7) && (atoi(optarg) >= 0)) log_level = atoi(optarg);
				else {
					fprintf(stderr, "  Error - Log level must be a value between 0 and 6.\n");
					exit(1);
				}
				break;
			case 'v':
				fprintf(stdout, "%s\n", PACKAGE_STRING);
				exit(0);
			case 'h':
			case '?':
				usage(my_argv[0]);
			default:
				break;
		}
	}

	/* initialization phase, all logging goes to stdout or stderr */
	fprintf(stdout, "\nhoneytrap v%s - Initializing.\n", VERSION);

	
	DEBUG_FPRINTF(stdout, "  Saving old working directory.\n");
	if (getcwd(old_cwd, 1000) == NULL) {
		DEBUG_FPRINTF(stderr, "  Warning - Unable to determine current working directory.\n");
		exit(1);
	}
	

	/* scan command line options to get config file name */
	optind = 1;
	while((option = getopt(my_argc, my_argv, OPTSTRING)) > 0) {
		switch(option) {
			case 'C':
				/* config file */
				conffile_name = (char *) realloc(conffile_name, strlen(optarg)+1);
				snprintf(conffile_name, strlen(optarg)+1, "%s\0", optarg);
				DEBUG_FPRINTF(stdout, "  Reading configuration from %s.\n", conffile_name);
				break;
			case 'D':
				/* don't daemonize */
				daemonize = 0;
				DEBUG_FPRINTF(stdout, "  Not daemonizing - staying in foreground.\n");
				break;
			default:
				break;
		}
	}
	/* process config file */
	if (parse_config_file() == -1) {
		fprintf(stderr, "  Error - Unable to parse configuration file.\n");
		exit(1);
	}


	/* scan command line options to get logfile name */
	optind = 1;
	while((option = getopt(my_argc, my_argv, OPTSTRING)) > 0) {
		switch(option) {
			case 'L':
				logfile_name = (char *) realloc(logfile_name, strlen(optarg)+1);
				snprintf(logfile_name, strlen(optarg)+1, "%s\0", optarg);
				break;
			default:
				break;
		}
	}


	/* process remaining options now */
	optind = 1;
	while((option = getopt(my_argc, my_argv, OPTSTRING)) > 0) {
		switch(option) {
			case 'i':
				dev = (char *) malloc(strlen(optarg)+1);
				strncpy(dev, optarg, strlen(optarg));
				dev[strlen(optarg)] = 0;

				/* get IPv4 address from interface */
				if (interface = get_if_data(AF_INET, dev)) {
					sa_in = (struct sockaddr_in *) interface->if_addr;
					if ( (sa = interface->if_addr) != NULL) {
						this_host = sa_in->sin_addr;
						DEBUG_FPRINTF(stdout, "  Interface %s has IP address %s\n",
							dev, inet_ntoa(this_host));

						/* need a 2-bytes offset for ip header from p2p devices */
						if (interface->if_flags & IFF_POINTOPOINT)
							DEBUG_FPRINTF(stdout, "  %s is a point-to-point device, using a 2 bytes offset.\n", dev);
						p2p_offset = (interface->if_flags & IFF_POINTOPOINT) ? 2 : 0;
						free_if_data(interface);
					}
				} else {
					fprintf(stderr, "  Error - Interface %s is down or no IPv4 address is configured.\n", optarg);
					exit(1);
				}
				break;
			case 'a':
				if ((ip_cmd_opt = gethostbyname(optarg)) == NULL) {
					fprintf(stderr, "  Error - Invalid hostname or ip address.\n");
					exit(1);
				}
				break;
			case 'p':
				promisc_mode = 1;
				break;
			case 'm':
				mirror_mode = 1;
				break;
			case 'l':
				conn_timeout = strtoul(optarg, NULL, 0);
				if((conn_timeout < 0) || (conn_timeout > 255)) {
					fprintf(stderr, "  Error - Listen timeout must be a value between 0 and 255.\n");
					exit(1);
				}
				break;
			case 'r':
				read_timeout = strtoul(optarg, NULL, 0);
				if((read_timeout < 0) || (read_timeout > 255)) {
					fprintf(stderr, "  Error - Read timeout must be a value between 0 and 255.\n");
					exit(1);
				}
				break;
			case 'u':
				if(pwd_entry = getpwnam(optarg)) {
					u_id = pwd_entry->pw_uid;
					user = (char *) strdup(optarg);
				} else {
					fprintf(stdout, "  Warning - Invalid user. Servers will run as root.\n");
					u_id = 0;
					user = (char *) strdup("root");
				}
				break;
			case 'g':
				if(grp_entry = getgrnam(optarg)) {
					g_id = grp_entry->gr_gid;
					group = (char *) strdup(optarg);
				} else {
					fprintf(stdout, "  Warning - Invalid group. Servers will run as root.\n");
					g_id = 0;
					group = (char *) strdup("root");
				}
				break;
			case 'P':
				/* pid file */
				pidfile_name = (char *) realloc(pidfile_name, strlen(optarg)+1);
				snprintf(pidfile_name, strlen(optarg)+1, "%s\0", optarg);
				DEBUG_FPRINTF(stdout, "  Pid file is %s.\n", pidfile_name);
				break;
			case 'D':
			case 'L':
			case 't':
			default:
				break;
		}
	}
	fprintf(stdout, "  Servers will run as user %s (%d).\n", user, u_id);
	fprintf(stdout, "  Servers will run as group %s (%d).\n", group, g_id);

	
	/* load plugins */
	load_plugins(plugin_dir);


	/* load default responses */
	if (response_dir) {
		if (load_default_responses(response_dir) == -1)
			fprintf(stdout, "  Warning - Unable to load default responses.\n");
	}


	if(promisc_mode) fprintf(stdout,"  Promiscuous mode enabled.\n");
	if(mirror_mode) fprintf(stdout,"  Promiscuous mode enabled.\n");


	/* assemble bpf expression from command line */
	if (optind < my_argc) {
		while (optind < my_argc) {
			if (!bpf_cmd_ext) {
				bpf_cmd_ext = (char *) strdup(my_argv[optind]);
			} else {
				if ((bpf_cmd_ext = (char *) realloc(bpf_cmd_ext,
					strlen(bpf_cmd_ext)+strlen(my_argv[optind])+2)) == NULL) {
					fprintf(stderr,
						"  Error - Unable to allocate memory: %s\n", strerror(errno));
					exit(1);
				}
				snprintf(bpf_cmd_ext+strlen(bpf_cmd_ext), strlen(my_argv[optind])+2,
					" %s\0", my_argv[optind]);
			}
			my_argv[optind++];
		}
		DEBUG_FPRINTF(stdout, "  Command line bpf expression is '%s'\n", bpf_cmd_ext); 
	}

	/* create berkeley packet filter string */
	if (bpf_filter_string) {
		free(bpf_filter_string);
		bpf_filter_string = NULL;
	}
	bpf_filter_string = (char *) strdup(create_bpf(bpf_cmd_ext, ip_cmd_opt, interface));
	free(ip_cmd_opt);


	/* open logfile */
	if((logfile_fd = open(logfile_name, EXCL_FILE_RW, 0644)) == -1) {
		fprintf(stderr, "  Error - Unable to open logfile.\n");
		exit(1);
	}
	fprintf(stdout, "  Logging to %s.\n", logfile_name);


	/* install signal handlers */
	install_signal_handlers();

	fprintf(stdout, "  Initialization completed\n");
	logmsg(LOG_ERR, 0, "\nhoneytrap v%s Copyright (C) 2005-2006 Tillmann Werner <tillmann.werner@gmx.de>\n", VERSION);
	return(1);
}


int parse_config_file(void) {
	char *recvline, *config_opt, *config_val;
	int config_file_fd, line_number;;
	FILE *config_file;

	line_number = 0;
	recvline = NULL;
	config_opt = NULL;
	config_val = NULL;


	if((config_file_fd = open(conffile_name, 0, 0640)) == -1) return(-1);

	/* reopen config file as stream to be able to use fgets() */
	if ((config_file = fdopen(config_file_fd, "r")) == NULL) {
		fprintf(stderr, "  Error - Unable to open configuration file: %s\n", strerror(errno));
		return(-1);
	}

	DEBUG_FPRINTF(stdout, "  Config file parser - File reopened as stream, ready to parse.\n");

	while(recvline = get_next_line(config_file)) {
		line_number++;
		
		/* ignore comments and blank lines */
		if ((*recvline == '#') || (*recvline == 0x0a) || (*recvline == 0x00) || (recvline == NULL)) continue;

		config_opt = recvline;
		if ((config_val = get_value(recvline, '=')) != NULL) {
			/* remove newlines */
			if ((config_val)[strlen(config_val)-1] == '\n') config_val[strlen(config_val)-1] = 0;
			DEBUG_FPRINTF(stdout, "  Config file parser - Line %u: %s=%s\n",
				line_number, config_opt, config_val);
		} else {
			if ((config_opt)[strlen(config_opt)-1] == '\n') config_opt[strlen(config_opt)-1] = 0;
			DEBUG_FPRINTF(stdout, "  Config file parser - Line %u: %s\n",
				line_number, config_opt);
			config_val = NULL;
		}
		if (process_config_option(config_opt, config_val) == -1) {
			fprintf(stderr, "  Error - Invalid line %u in configuration file.\n", line_number); 
			return(-1);
		}
	}
	DEBUG_FPRINTF(stdout, "  Config file parser - %u lines successfully parsed.\n", line_number); 

	if (fclose(config_file) == EOF) fprintf(stderr, "  Error - Unable to close configuration file: %s\n",
		strerror(errno));
	
	return(0);
}

int process_config_option(char *opt, char* val) {
	char *index, *low_opt;
	struct passwd *pwd_entry;
	struct group *grp_entry;

	/* turn keywords into lowercase */
	low_opt = (char *) strdup(opt);
	index = low_opt;
	while(*index != '\0') tolower(index++);

	
	if (OPT_IS("mirror")) {
		mirror_mode = 1;
		DEBUG_FPRINTF(stdout, "  Activating mirror mode.\n");
	} else if (OPT_IS("promisc")) {
		promisc_mode = 1;
		DEBUG_FPRINTF(stdout, "  Activating promiscuous mode.\n");
	} else if (OPT_IS("pidfile")) {
		pidfile_name = (char *) realloc(pidfile_name, strlen(val)+1);
		snprintf(pidfile_name, strlen(val)+1, "%s\0", val);
		DEBUG_FPRINTF(stdout, "  Pid file %s.\n", val);
	} else if (OPT_IS("logfile")) {
		logfile_name = (char *) realloc(logfile_name, strlen(val)+1);
		snprintf(logfile_name, strlen(val)+1, "%s\0", val);
		DEBUG_FPRINTF(stdout, "  Logfile is %s.\n", val);
	} else if (OPT_IS("dbfile")) {
		dbfile_name = (char *) realloc(dbfile_name, strlen(val)+1);
		snprintf(dbfile_name, strlen(val)+1, "%s\0", val);
		DEBUG_FPRINTF(stdout, "  Databse file is %s.\n", val);
	} else if (OPT_IS("response_dir")) {
		response_dir = (char *) realloc(response_dir, strlen(val)+1);
		snprintf(response_dir, strlen(val)+1, "%s\0", val);
		DEBUG_FPRINTF(stdout, "  Loading default responses from %s.\n", val);
	} else if (OPT_IS("plugin_dir")) {
		plugin_dir = (char *) realloc(plugin_dir, strlen(val)+1);
		snprintf(plugin_dir, strlen(val)+1, "%s\0", val);
		DEBUG_FPRINTF(stdout, "  Loading plugins from %s.\n", val);
	} else if (OPT_IS("attacks_dir")) {
		attacks_dir = (char *) realloc(attacks_dir, strlen(val)+1);
		snprintf(attacks_dir, strlen(val)+1, "%s\0", val);
		DEBUG_FPRINTF(stdout, "  Storing attack strings in %s.\n", val);
	} else if (OPT_IS("dlsave_dir")) {
		dlsave_dir = (char *) realloc(dlsave_dir, strlen(val)+1);
		snprintf(dlsave_dir, strlen(val)+1, "%s\0", val);
		DEBUG_FPRINTF(stdout, "  Storing downloaded files in %s.\n", val);
	} else if (OPT_IS("user")) {
		if(pwd_entry = getpwnam(val)) {
			u_id = pwd_entry->pw_uid;
			user = (char *) strdup(val);
		} else fprintf(stdout, "  Invalid user in configuration file..\n");
	} else if (OPT_IS("group")) {
		if(grp_entry = getgrnam(val)) {
			g_id = grp_entry->gr_gid;
			group = (char *) strdup(val);
		} else fprintf(stdout, "  Invalid group in configuration file..\n");
	} else {
		fprintf(stderr, "  Error - Invalid keyword '%s' in configuration file.\n", opt);
		return(-1);
	}

	free(low_opt);
	
	return(0);
}


char *create_bpf(char *bpf_cmd_ext, struct hostent *ip_cmd_opt, struct if_data *interface) {
	char *bpf_filter_string, errbuf[PCAP_ERRBUF_SIZE];
	struct in_addr netaddr, netmask;

	bpf_filter_string = NULL;

	/* determine ip address of device */
	if(dev == NULL) {
		dev = (char *) malloc(4);
		dev = "any";
	} else if ((interface) && (strcmp(dev, "any") != 0)) {
		/* lookup net address and netmask for interface */
		pcap_lookupnet(dev, &net, &mask, errbuf);
		netaddr.s_addr = net;
		netmask.s_addr = mask;
	}

	/* assemble filter string */
	if (!(bpf_filter_string = (char *)malloc(30))) {
		fprintf(stderr, "  Error - Not enough memory for filter string.\n");
		exit(1);
	} else {
		snprintf(bpf_filter_string, 29, "tcp[tcpflags] & tcp-rst != 0\0");
		
		if ((interface) && (strcmp(dev, "any") != 0)) {
			/* add local ip address (from interface or -a command line option) to filter string */
				bpf_filter_string = (char *) realloc(bpf_filter_string, strlen(bpf_filter_string)+30);
			strncat(bpf_filter_string, " and src host \0", 14);
			if (ip_cmd_opt) {
				strncat(bpf_filter_string, (char*) inet_ntoa(*(struct in_addr*)ip_cmd_opt->h_addr_list[0]),
					strlen((char *) inet_ntoa(*(struct in_addr*)ip_cmd_opt->h_addr_list[0])));
			} else {
				strncat(bpf_filter_string, (char *) inet_ntoa(this_host),
					strlen((char *) inet_ntoa(this_host)));
			}
			strncat(bpf_filter_string, "\0", 1);
		}

		/* add bpf expression from command line */
		if (bpf_cmd_ext) {
			if (!(bpf_filter_string = (char *)realloc(bpf_filter_string,
				strlen(bpf_filter_string)+strlen(bpf_cmd_ext)+8))) {
				fprintf(stderr, "  Error - Unable to allocate memory: %s\n", strerror(errno));
				exit(1);
			}	
			snprintf(bpf_filter_string+strlen(bpf_filter_string), strlen(bpf_cmd_ext)+8,
				" and (%s)\0", bpf_cmd_ext);
		}
	}
	fprintf(stdout, "  BPF string is '%s'.\n", bpf_filter_string);

	return(bpf_filter_string);
}
