/* tcpserver.c
 * Copyright (C) 2005-2006 Tillmann Werner <tillmann.werner@gmx.de>
 *
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without
 * modifications, as long as this notice is preserved.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <errno.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <string.h>

#include "honeytrap.h"
#include "logging.h"
#include "tcpserver.h"
#include "mirror.h"
#include "plughook.h"

u_char buffer[BUFSIZ], *attack_string;

int drop_privileges(void) {
	/* set gid first, it might not be permitted as unprivileged user */
	if(setgid(g_id) != 0) {
		logmsg(LOG_WARN, 1, "Warning - unable to set group id for server to %d.\n", g_id);
		return(1);
	}
	if(setuid(u_id) != 0) {
		logmsg(LOG_WARN, 1, "Warning - unable to set user id for server to %d.\n", u_id);
		return(1);
	}
	return(0);
}


void start_tcp_server(u_char *args, const struct pcap_pkthdr *header, const u_char * packet) {
    pid_t pid;
    int listenfd, mirror_sock_fd, connection_fd, max_read_fd, disconnect, sock_option, i,
	client_addr_len, bytes_read, bytes_sent, total_bytes, select_return;
    struct sockaddr_in client_addr, server_addr;
    struct timeval c_timeout, r_timeout;
    fd_set rfds;
    

    ip = (struct ip_header *) (packet + ETHER_HDRLEN + p2p_offset);
    tcp = (struct tcp_header *) (packet + ETHER_HDRLEN + ip->ip_hl * 4 + p2p_offset);

    attack_string	= NULL;
    select_return	= -1;


    /* new connections are welcome */
    if (ntohl(tcp->th_seqno) != 0) return;

    logmsg(LOG_NOISY, 1, "-> %u\t  Caught reset packet to %s, forking server process.\n",
	    (u_int16_t) ntohs(tcp->th_sport), inet_ntoa(ip->ip_dst));

    /* fork server process */
    if ((pid = fork()) == 0) {
	if (!(listenfd = socket(AF_INET, SOCK_STREAM, 0))) {
	    logmsg(LOG_ERR, 1, "Error - socket() call failed: %s\n", strerror(errno));
	    exit(1);
	}

	sock_option = 1;
	if (setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &sock_option, sizeof(sock_option)) < 0)
	logmsg(LOG_WARN, 1, "Warning - Unable to set SO_REUSEADDR on listening socket.\n");

	bzero((char *) &server_addr, sizeof(server_addr));
	server_addr.sin_family = AF_INET;
	server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	server_addr.sin_port = tcp->th_sport;
	if ((bind(listenfd, (struct sockaddr *) &server_addr, sizeof(server_addr))) < 0) {
	    /* bind failures can happen because of resetted connections of other apps */
	    logmsg(LOG_ERR, 1, "Error - bind() call failed for port %u/tcp: %s\n",
		(u_int16_t) ntohs(tcp->th_sport), strerror(errno));
	    close(listenfd);
	    exit(1);
	}   

	drop_privileges(); 
	logmsg(LOG_DEBUG, 1, "Server is now running with user id %d and group id %d.\n", getuid(), getgid());
		  

	if ((listen(listenfd, 0)) < 0) {
	    logmsg(LOG_ERR, 1, "Error - listen() call failed: %s\n", strerror(errno));
	    close(listenfd);
	    exit(1);
	}
	
	client_addr_len = sizeof(client_addr);

	/* wait for incoming connections */
	for (;;) {
	    FD_ZERO(&rfds);
	    FD_SET(listenfd, &rfds);
	    c_timeout.tv_sec = conn_timeout;
	    c_timeout.tv_usec = 0;
	    select_return = select(listenfd + 1, &rfds, NULL, NULL, &c_timeout);

	    if (select_return < 0) {
		if (errno != EINTR) {
		    logmsg(LOG_ERR, 1, "Error - select() call failed: %s\n", strerror(errno));
		    exit(1);
		}
	    } else if (select_return == 0) {
		/* timeout */
		logmsg(LOG_NOISY, 1, "-> %u\t  No incoming connection for %u seconds - server terminated.\n",
		    (u_int16_t) ntohs(tcp->th_sport), conn_timeout);
		exit(0);
	    } else {
		if (FD_ISSET(listenfd, &rfds)) {
		    logmsg(LOG_NOISY, 1, "   %u\t  Connection request from %s:%u.\n",
			(u_int16_t) ntohs(tcp->th_sport), inet_ntoa(ip->ip_dst), (u_int16_t) ntohs(tcp->th_dport));

		    /* try establish mirror connection back to the client */
		    if (mirror_mode) {
			logmsg(LOG_DEBUG, 1, "<> %u\t  Requesting mirror connection to %s:%u.\n",
			    (u_int16_t) ntohs(tcp->th_sport), inet_ntoa(ip->ip_dst), (u_int16_t) ntohs(tcp->th_sport));
			if ((mirror_sock_fd = mirror_connect((struct in_addr) ip->ip_dst, tcp->th_sport)) == -1) {
			    logmsg(LOG_INFO, 1, "<> %u\t  Mirror connection rejected, falling back to normal mode.\n",
				(u_int16_t) ntohs(tcp->th_sport));
			    mirror_mode = 0;
			} else logmsg(LOG_NOTICE, 1, "<> %u\t  Mirror connection to %s:%u established.\n",
			    (u_int16_t) ntohs(tcp->th_sport), inet_ntoa(ip->ip_dst), (u_int16_t) ntohs(tcp->th_sport));
		    }

		    /* accept connection request */
		    if ((connection_fd = accept(listenfd, (struct sockaddr *) &client_addr, &client_addr_len)) < 0) {
			if (errno == EINTR) break;
			else {
			    logmsg(LOG_ERR, 1, "Error - accept() call failed: %s\n", strerror(errno));
			    close(mirror_sock_fd);
			    exit(1);
			}
		    } else {
			/* accept successful, fork process and handle connection */
			logmsg(LOG_NOTICE, 1, "   %u\t  Connection accepted from %s:%u.\n",
			    (u_int16_t) ntohs(tcp->th_sport), inet_ntoa(ip->ip_dst), (u_int16_t) ntohs(tcp->th_dport));
			if ((pid = fork()) == 0) {
			    /* close listening socket in child */
			    close(listenfd);
			    disconnect = 0;
			    total_bytes = 0;

			    init_attack(connection_fd, ip, tcp);	/* initialize attack record */

			    if (mirror_mode) {
				logmsg(LOG_DEBUG, 1, "   %u\t  Handling connection from %s:%u in mirrored mode.\n",
				    (u_int16_t) ntohs(tcp->th_sport),
				    inet_ntoa(ip->ip_dst), (u_int16_t) ntohs(tcp->th_dport));
				handle_connection_mirrored(connection_fd, mirror_sock_fd, (u_int16_t) ntohs(tcp->th_sport),
					(u_int16_t) ntohs(tcp->th_dport), ip->ip_dst, m_read_timeout, read_timeout);
				exit(0);
			    } else {
				logmsg(LOG_DEBUG, 1, "   %u\t  Handling connection from %s:%u in normal mode.\n",
				    (u_int16_t) ntohs(tcp->th_sport),
				    inet_ntoa(ip->ip_dst), (u_int16_t) ntohs(tcp->th_dport));
				handle_connection_normal(connection_fd, (u_int16_t) ntohs(tcp->th_sport), read_timeout);
				exit(0);
			    }

			} else if (pid == -1) logmsg(LOG_ERR, 1, "Error - forking connection handler failed.\n");
			close(mirror_sock_fd);
			close(connection_fd);
		    } /* connection accepted */
		} /* FD_ISSET - incoming connection */
	    } /* select return for listenfd */	
	} /* for - incoming connections */
	/* end server process */
    } /* fork - server process */
    else if (pid == -1) logmsg(LOG_ERR, 1, "Error - forking server process failed.\n");
    return;
}


/* handle connection in mirror mode - send data mirrored from the connecting client */
int handle_connection_mirrored(int connection_fd, int mirror_sock_fd, u_int16_t dport, u_int16_t sport, struct in_addr ipaddr, u_char timeout, u_char fb_timeout) {
    fd_set rfds;
    struct timeval r_timeout;
    int disconnect, bytes_read, bytes_sent, total_bytes, total_from_mirror, max_read_fd, retval;
    u_char *mirror_string;

    disconnect	= 0;
    total_bytes	= 0;
    total_from_mirror = 0;
    mirror_string = NULL;

    /* read data from sockets */
    for (;;) {
	FD_ZERO(&rfds);
	FD_SET(connection_fd, &rfds);
	FD_SET(mirror_sock_fd, &rfds);

	max_read_fd = (mirror_sock_fd > connection_fd ? mirror_sock_fd : connection_fd);

	r_timeout.tv_sec = (u_char) timeout;
	r_timeout.tv_usec = 0;
	if ((select(connection_fd + 1, &rfds, NULL, NULL, &r_timeout) < 0) && (errno != EINTR)) {
	    logmsg(LOG_ERR, 1, "Error - %u\t  Select failed - connection handler terminated.\n", dport);
	    close(mirror_sock_fd);
	    return(process_data(attack_string, total_bytes, mirror_string, total_from_mirror, dport));
	}

	if (FD_ISSET(mirror_sock_fd, &rfds)) {
	/* read data and mirror it to client connection */
	    bytes_read = 0;
	    if ((retval =
		tcpcopy(connection_fd, mirror_sock_fd, &mirror_string, total_from_mirror, &bytes_read, &bytes_sent)) > 0) {
		logmsg(LOG_INFO, 1, "<> %u\t* %u (of %u) bytes copied from mirror connection to %s:%u.\n",
			dport, bytes_sent, bytes_read, inet_ntoa(ipaddr), sport);
		total_from_mirror += bytes_read;
	    } else if (retval == 0) {
		/* remote host closed mirror connection */
		logmsg(LOG_INFO, 1, "   %u\t  Mirror connection closed by foreign host.\n", dport);
	    	return(process_data(attack_string, total_bytes, mirror_string, total_from_mirror, dport));
	    } else {
		/* tcpcopy error */
		logmsg(LOG_INFO, 1, "   %u\t  Error - Unable to mirror data to client connection.\n", dport);
		if (close(mirror_sock_fd) == -1) logmsg(LOG_ERR, 1, "Error - Unable to close mirror sockt.\n");
			else logmsg(LOG_NOISY, 1, "<> %u\t  Mirror connection closed.\n", dport, inet_ntoa(ipaddr), dport);
	    	return(process_data(attack_string, total_bytes, mirror_string, total_from_mirror, dport));
	    }
	} else if (FD_ISSET(connection_fd, &rfds)) {
	/* read data and mirror it to mirror connection */
	    bytes_read = 0;
	    if ((retval= tcpcopy(mirror_sock_fd, connection_fd, &attack_string, total_bytes, &bytes_read, &bytes_sent))>0) {
		logmsg(LOG_INFO, 1, "<> %u\t* %u (of %u) bytes copied from client connection to %s:%u.\n",
			dport, bytes_sent, bytes_read, inet_ntoa(ipaddr), dport);
		total_bytes += bytes_read;
	    } else if (retval == 0) {
		/* remote host closed client connection */
		logmsg(LOG_INFO, 1, "   %u\t  Connection closed by foreign host.\n", dport);
		close(mirror_sock_fd);
		logmsg(LOG_NOISY, 1, "<> %u\t  Mirror connection closed.\n", dport, inet_ntoa(ipaddr), dport);
	    	return(process_data(attack_string, total_bytes, mirror_string, total_from_mirror, dport));
	    } else {
		/* tcpcopy error */
		logmsg(LOG_INFO, 1, "   %u\t  Error - Unable to mirror data to mirror connection.\n", dport);
	    	return(process_data(attack_string, total_bytes, mirror_string, total_from_mirror, dport));
	    }
	} else {
		/* select() timed out, close connections */
		logmsg(LOG_INFO, 1, "   %u\t  Mirror connection timed out, closing connections.\n", dport);
		close(mirror_sock_fd);
		return(0);
	}
    }
}


/* handle connection in normal mode - respond with default answers */
int handle_connection_normal(int connection_fd, u_int16_t port, u_char timeout) {
    fd_set rfds;
    struct timeval r_timeout;
    int disconnect, bytes_read, total_bytes;
	    
    total_bytes	= 0;
    disconnect	= 0;

    /* read data from sockets */
    for (;;) {
	FD_ZERO(&rfds);
	FD_SET(connection_fd, &rfds);

	if (((u_int16_t) ntohs(port) == 80)
	    || ((u_int16_t) ntohs(port) == 135)
	    || ((u_int16_t) ntohs(port) == 139)
	    || ((u_int16_t) ntohs(port) == 445)
	    || ((u_int16_t) ntohs(port) == 1433) 
	    || ((u_int16_t) ntohs(port) == 4444)) {
	    conn_timeout = 2;
	    disconnect++;
	} else disconnect += 2;

	r_timeout.tv_sec = (u_char) timeout;
	r_timeout.tv_usec = 0;
	if ((select(connection_fd + 1, &rfds, NULL, NULL, &r_timeout) < 0) && (errno != EINTR)) {
	    logmsg(LOG_ERR, 1, "Error - %u\t  Select failed - connection handler terminated.\n", port);
	    return(process_data(attack_string, total_bytes, NULL, 0, port));
	}

	/* handle data on server connection */
	if (FD_ISSET(connection_fd, &rfds)) {
	    if ((bytes_read = read(connection_fd, buffer, sizeof(buffer))) > 0) {
		logmsg(LOG_INFO, 1, "   %u\t* %d bytes read.\n", port, bytes_read);
		total_bytes += bytes_read;
		if (!(attack_string = (u_char *) realloc(attack_string, total_bytes))) {
		    logmsg(LOG_ERR, 1, "Error - %u\t  Reallocating buffer size failed.\n", port);
		    free(attack_string);
		    return(-1);
		}
		memcpy(attack_string + total_bytes - bytes_read, buffer, bytes_read);
    		disconnect = 0;
	    } else {
		logmsg(LOG_INFO, 1, "   %u\t  Connection closed by foreign host.\n", port);

		/* process attack string */
		return(process_data(attack_string, total_bytes, NULL, 0, port));
	    }
	} else {
	    /* no data available, select timed out */
	    if (disconnect > 5) {
		logmsg(LOG_INFO, 1, "   %u\t  Timeout expired, closing connection.\n", port);

		/* process attack string */
		return(process_data(attack_string, total_bytes, NULL, 0, port));
	    } else {
		if ((!send_default_response(connection_fd, port, read_timeout)) == -1) {
		    logmsg(LOG_ERR, 1, "   %u\t  Response failed. Closing socket.\n");
		    return(process_data(attack_string, total_bytes, NULL, 0, port));
		}
	    } /* disconnect check */
	} /* FD_ISSET */
    } /* for */
}


/* process attack - call plugins registered for hook 'process_attack' */
int process_data(u_char *a_data, u_int32_t a_size, u_char *m_data, u_int32_t m_size, u_int16_t port) {

	/* save end time and payload data in attack record */
	time(&attack.end_time);
	/* attack string */
	attack.a_conn.payload.size = a_size;
	if (a_size) {
		attack.a_conn.payload.data = (char *) malloc(a_size);
		memcpy(attack.a_conn.payload.data, a_data, a_size);
	}
	memcpy(attack.a_conn.payload.chksum, (char*)mem_md5sum(attack.a_conn.payload.data, attack.a_conn.payload.size), 33);
	/* mirror string */
	attack.m_conn.payload.size = m_size;
	if (m_size) {
		attack.m_conn.payload.data = (char *) malloc(m_size);
		memcpy(attack.m_conn.payload.data, m_data, m_size);
	}
	memcpy((char *) &(attack.m_conn.payload.chksum),
		(char *) mem_md5sum(attack.m_conn.payload.data, attack.m_conn.payload.size), 32);


	if (!a_size) {
		logmsg(LOG_NOTICE, 1, " * %u\t  No bytes received from %s:%u.\n",
		(u_int16_t) ntohs(tcp->th_sport), inet_ntoa(ip->ip_dst), (u_int16_t) ntohs(tcp->th_dport));
	} else {
		logmsg(LOG_NOTICE, 1, " * %u\t  %d bytes attack string from %s:%u.\n", (u_int16_t) ntohs(tcp->th_sport),
			a_size, inet_ntoa(ip->ip_dst), (u_int16_t) ntohs(tcp->th_dport));
	}

	/* call plugins */
	/* do calls even if no data received, i.e. to update connection statistics */
	plughook_process_attack(attack);


	/* parse attack string for ftp, http, ... URIs and download them */
//	get_online_ressources(a_data, a_size);

	return(1);
}


void init_attack(int fd, const struct ip_header *ip, const struct tcp_header *tcp) {
	/* clean attack record */
	if (attack.a_conn.payload.data) free(attack.a_conn.payload.data);	/* free attack data */
	if (attack.m_conn.payload.data) free(attack.m_conn.payload.data);	/* free mirror data */
	if (attack.p_conn.payload.data) free(attack.p_conn.payload.data);	/* free proxy data */
	attack.a_conn.payload.data == NULL;
	attack.m_conn.payload.data == NULL;
	attack.p_conn.payload.data == NULL;
	bzero(&attack.a_conn.payload, sizeof(struct s_payload));
	bzero(&attack.m_conn.payload, sizeof(struct s_payload));
	bzero(&attack.p_conn.payload, sizeof(struct s_payload));
	bzero(&attack.a_conn, sizeof(struct s_conn));
	bzero(&attack.m_conn, sizeof(struct s_conn));
	bzero(&attack.p_conn, sizeof(struct s_conn));
	bzero(&attack, sizeof(struct s_attack));

	/* store attack connection data in attack record */
	attack.a_conn.r_addr	= ip->ip_src;
	attack.a_conn.l_addr	= ip->ip_dst;
	attack.a_conn.r_port	= ntohs(tcp->th_dport);
	attack.a_conn.l_port	= ntohs(tcp->th_sport);
	time(&attack.start_time);

	return;
}


/* tcpcopy - reads data from one connection and writes it to another *
 * also store read data in 'save_string' at position 'offset' */
int tcpcopy(int to_fd, int from_fd, u_char **save_string, u_int32_t offset, int *bytes_read, int *bytes_sent) {
	/* read from from_sock_fd */
	if ((*bytes_read = read(from_fd, buffer, sizeof(buffer))) > 0) {
		/* write read bytes to save_string at offset */
		if (!(*save_string = (u_char *) realloc(*save_string, offset+(*bytes_read)))) {
			logmsg(LOG_ERR, 1, "Error - Unable to allocate memory: %s\n", strerror(errno));
			close(from_fd);
			close(to_fd);
			return(-1);
		}
		memcpy(*save_string + offset, buffer, (*bytes_read));

		/* write read bytes to to_sock_fd */
		*bytes_sent = 0;
		if ((*bytes_sent = write(to_fd, buffer, *bytes_read)) == -1) {
			logmsg(LOG_ERR, 1, "Unable to tcpcopy() %u bytes to target connection.\n", *bytes_read);
			close(from_fd);
			close(to_fd);
			return(-1);
		}
		return(*bytes_sent);
	}
	return(0);
}
